#include "circularvoting.h"

#include <image_io.h>
#include <scalespace.h>
#include <bip.h>

using namespace stdi;
using namespace voting;
using namespace image_io;
using namespace bip;
using namespace scalespace;

#include <cmath>

int WPoint2D::nx = 0;
int WPoint2D::ny = 0;

int Cone2D::nx = 0;
int Cone2D::ny = 0;

int ConePlane2D::nx = 0;
int ConePlane2D::ny = 0;

// 2D cone
Cone2D::Cone2D()  : dx(1), dy(0) { }
void Cone2D::setDirection(float xx, float yy)
{ dx = xx; dy = yy; }

// Static constants
const float CircularVoting::epsilon(0.001);
const float CircularVoting::pi = 3.1415927;

CircularVoting::CircularVoting(){

  _min_grad = 1;
  
  _hmin = 1;
  _hmax = 20;
  _radius = 10;
  
  _reg_type = AllReg;
  
  ntheta = 256;
  delta_theta = 2*pi/ntheta;
  
  _zc_only = false;
  
  _save_history = false;
  
  _abs_thresh = true;
}


void CircularVoting::setParams(RegionType reg_type, 
			       int hmin, 
			       int hmax, 
			       int r, 
			       float min_grad,
			       float threshold, 
			       float scale,
			       WeightingMethod weighting_method,
			       float weighting_param,
			       bool zc_only,
			       bool abs_thresh){
  
  _abs_thresh = abs_thresh;
  _reg_type = reg_type;
  _hmin = max<int>(hmin, 1); 
  _hmax = max<int>(hmax, _hmin);
  //_radius = max<int>(r, 1);
  _radius = max<int>(r, 0);
  _min_grad = min_grad;
  
  if( !_abs_thresh )
    _threshold = max<float>(0.05, min<float>(1, threshold));
  else
    _threshold = threshold;

  //cout<<"Set ParamsThreshold "<<threshold<<endl;
  _scale = max<float>(0, scale);
  
  _weighting_method = weighting_method;
  _weighting_param = weighting_param;
  
  _zc_only = zc_only;
  
}

void CircularVoting::saveHistory(bool yn){
  _save_history = yn;
}




void CircularVoting::compute(const Image& I)
{
  
  //cout<<"Executing local version of circular voting "<<" Thresh :"<<_threshold<<endl;
  
  nx = I.nx();
  ny = I.ny();
  npix = nx*ny;
  
  _voting_points = vector<VPoint2D>();
  _voting_points.reserve(npix/2);
    
  const float reg_sign = _reg_type==DarkReg ? -1 : 1;
  
  FloatImage Dx = dX(I, _scale);
  FloatImage Dy = dY(I, _scale);
  FloatImage G(nx, ny);
  for(int i=0; i<npix; i++) {
    Dx(i) *= reg_sign;
    Dy(i) *= reg_sign;
    float g = sqrt(Dx(i)*Dx(i)+Dy(i)*Dy(i));
    G(i) = g;
    if (g>epsilon) {
      Dx(i) /= g;
      Dy(i) /= g;
    }
  }
  
  float maxgrad = G.max();
  if (maxgrad>epsilon) 
    for(int i=0; i<npix; i++) {
      if (G(i)<_min_grad)
	G(i) = 0;
      else if( !_abs_thresh )
	G(i) /= maxgrad;
    }
  
  
  if (_zc_only) {
    ByteImage ZC = edge(features(I, _scale, BrightZC));    
    for(int i=0; i<npix; i++) 
      if (!ZC(i))
	G(i) = 0;
  }
  
  // Weighting   
  if (_weighting_method==PowMethod && fabs(_weighting_param-1)>epsilon) 
    for(int i=0; i<npix; i++) 
      if (G(i)>epsilon)
	G(i) = pow(G(i), _weighting_param);
      else if (_weighting_method==ExpMethod) 
	for(int i=0; i<npix; i++) 
	  if (G(i)>epsilon)
	    G(i) = exp( G(i) *  _weighting_param );
  
  
  const int nx_1 = nx-1;
  const int ny_1 = ny-1;
  
  bw = sqrt((double)(_radius*_radius + _hmax*_hmax)) + 3;
  const int bw2 = 2*bw;
  
  int indx;
  VPoint2D vp;
  int x, y, k;
  for(y = 0; y < ny; y++)
    for(x = 0; x < nx; x++) {      
      k = I.pos(x, y);
      if (G(k) > epsilon) {
	if ((indx=computeAngleIndex(Dx(k), Dy(k))) >= 0 ) {
	  vp.x = x + bw;
	  vp.y = y + bw;
	  vp.mag = G(k); 
	  vp.angIndex = indx;
	  _voting_points.push_back(vp);
	}	
      }
    }
  
  //cout << "#Voting points = " << _voting_points.size() << endl;
  
  nx += bw2;
  ny += bw2;
  npix = nx*ny;
    
  WPoint2D::setImageSize(nx, ny);
  ConePlane2D::setImageSize(nx, ny);
  Cone2D::setImageSize(nx, ny);
  
  vote();
  
  // Switch back to the original size
  nx -= bw2;
  ny -= bw2;
  npix = nx*ny;
  _vote = FloatImage(nx, ny);
  
  float val, maxval=0;
  
  for(y = 0; y < ny; y++)
    for(x = 0; x < nx; x++) {
      val = _sum(x+bw, y+bw);
      _vote(x, y) = val;
      maxval = max<float>(maxval, val);
    }
  //  return;
  int nc;
  //cout<<"Voting Threshold "<<_threshold<<" max Value "<<maxval<<" thresh * maxval:" <<_threshold * maxval<<endl;
  // ByteImage bIm = threshold(_vote,_threshold*maxval) ;
  // ByteImage bIm = threshold(_vote,_threshold) ;
  ByteImage bIm(_vote.nx(),_vote.ny(),_vote.nz());
  for (int count=0;count<bIm.npix();count++){
    if( _abs_thresh )
      bIm.set(count,_vote.pixel(count) > _threshold?255:0 );
    else
      bIm.set(count,_vote.pixel(count) > _threshold*maxval?255:0 );
  }
   
  ShortImage label = regions(bIm, nc) ;
  vector<RegionFeature> rf = regionFeatures(label, nc, _vote);
  _vote.reset(0);
  _center_weights.clear();
  for(vector<RegionFeature>::iterator it=rf.begin(); it!=rf.end(); it++) {
    if (_vote.inImage(it->xc, it->yc)) {
      _vote(it->xc, it->yc) = it->mean;
      _centers.push_back(Point( it->xc, it->yc ));
      _center_weights.push_back(it->mean);
    }   
  }
  
  for(vector<VPoint2D>::iterator it=_voting_points.begin(); it!=_voting_points.end(); it++) {
    it->x -= bw;
    it->y -= bw;
    it->xc -= bw;
    it->yc -= bw;
  }
  
}



CircularVoting::~CircularVoting()
{ }

void CircularVoting::setPrefix(const string& p)
{
  _prefix = p;
}

int CircularVoting::computeAngleIndex(float dx, float dy) const 
{
  float a = acos(dx);
  if (dy<0) {
    a = 2*pi-a;
  }
  
  int indx_theta = round_double(a/delta_theta);
  return max(0, min(ntheta-1, indx_theta));
}


FloatImage CircularVoting::votingImage()
{  return _vote;  }

const PointArray& CircularVoting::centers() const
{ return _centers; }

const vector<float>& CircularVoting::center_weights() const
{ return _center_weights; }


void Cone2D::vote(float* p, const VPoint2D& vp)
{
  iterator from = begin();
  iterator to = end();     
  for(iterator it = from; it != to; it++) 
    for(value_type::iterator it1=it->begin(); it1!=it->end(); it1++) {    
      p[it1->d] += vp.mag * it1->w;
    }      
}

#if 0

void Cone2D::vote(float* p, float* pmask, const VPoint2D& vp)
{
  float maxvote, m;  
  iterator it, from, to;
  value_type::iterator it1, it_end;
  
  maxvote = 0;  
  from = begin();
  to = end();
    
  for(it = from; it != to; it++) {	
    it1=it->begin();
    it->center_weight = pmask[it1->d];;
    maxvote = max(maxvote, it->center_weight);
  }
  
  if (maxvote<CircularVoting::epsilon)
    return;
    
  for(it = from; it != to; it++) {	
    m = vp.mag * it->center_weight / maxvote;
    it_end = it->end();    
    for(it1=it->begin(); it1!=it_end; it1++) {	 
      p[it1->d] += m * it1->w;	  
    }
  }
}

#endif

#if 1 // new 

void Cone2D::vote(float* p, float* pmask, const VPoint2D& vp)
{
  float sumvote, m;  
  iterator it, from, to;
  value_type::iterator it1, it_end;
  
  sumvote = 0;  
  from = begin();
  to = end();
  
  for(it = from; it != to; it++) {	
    it_end = it->end();    
    for(it1=it->begin(); it1!=it_end; it1++) {	 
      sumvote += pmask[it1->d];  
    }
  }
  
  if (sumvote<CircularVoting::epsilon)
    return;
  
  for(it = from; it != to; it++) {	
    m = vp.mag  / sumvote;
    it_end = it->end();    
    for(it1=it->begin(); it1!=it_end; it1++) {	 
      p[it1->d] += m * pmask[it1->d];	  
    }
  }
}

#endif

void CircularVoting::getMaxDirection(VPoint2D& vp)
{
  
  Cone2D& cone = _cones[vp.angIndex];
  
  Cone2D::iterator from = cone.begin();
  Cone2D::iterator to = cone.end();
  ConePlane2D::iterator max_point;
  
  float maxvote = 0;
    
  float* p = _sum.ptr() + vp.pos;
  for(Cone2D::iterator it = from; it != to; it++) {
    for(ConePlane2D::iterator it1=it->begin(); it1!=it->end(); it1++)
      if ( p[it1->d]>maxvote) {
	maxvote = max(maxvote, p[it1->d]);
	max_point = it1;
      }    
    } 
  
  if (maxvote<epsilon)
    return;
  
  float dx = max_point->x;
  float dy = max_point->y;
  
  float r = sqrt(dx*dx+dy*dy);
  if (r>epsilon) {
    vp.xc = vp.x + max_point->x;
    vp.yc = vp.y + max_point->y;
    vp.angIndex = computeAngleIndex(dx/r, dy/r);
  }
    
}


void CircularVoting::computeCones(int hmin, int hmax, int radius)
{
  _cones = vector<Cone2D>(ntheta);
  
  Cone2D cone;
  cone.reserve(2*hmax);
  
  cone.setDirection(1, 0);
    
  for(int x=hmin; x<=hmax; x++) {
    cone.push_back(ConePlane2D(1, WPoint2D(x, 0, 1)));
  }
  
  for(int x=hmin; x<=hmax; x++) {
    
    int rmax = round_double(x * radius / hmax);
    for(int y=-rmax; y<=rmax; y++)   if (y!=0) {
      
      float R = sqrt((double)(x*x+y*y));
      float a = fabs((float)x)/R;
      
      WPoint2D wp;
      wp.x = x; 
      wp.y = y; 
      wp.w = 1;// a;
      int n = round_double(R) - hmin;
      if (n>=0 && n<cone.size())
	cone[n].push_back(wp);
    }
    
  }
  
  if ( _reg_type==AllReg) {
    
    Cone2D cone1;
    cone1.setDirection(1, 0);
    
    for(int x=hmin; x<=hmax; x++) {
      cone1.push_back(ConePlane2D(1, WPoint2D(-x, 0, 1)));
    }
    
    for(int x=hmin; x<=hmax; x++) {
      
      int rmax = round_double(x * radius / hmax);
      for(int y=-rmax; y<=rmax; y++)   if (y!=0) {
	
	float R = sqrt((double)(x*x+y*y));
	float a = fabs((float)x)/R;
	
	WPoint2D wp;
	wp.x = -x; 
	wp.y = y; 
	wp.w = 1;// a;
	int n = round_double(R) - hmin;
	if (n>=0 && n<cone1.size())
	  cone1[n].push_back(wp);
      }
    }
    
    for(int i=0; i<cone1.size(); i++)
      cone.push_back(cone1[i]);
  }
  
  
  cone.init();
    
  for(int i=0; i<ntheta; i++) {
    
    float theta = i*delta_theta;
	
    Cone2D rcone = cone; // rotated cone
	
    const float cos_theta = cos(theta);
    const float sin_theta = sin(theta);
    
    for(Cone2D::iterator it1=rcone.begin(); it1!=rcone.end(); it1++)
      for(ConePlane2D::iterator it2=it1->begin(); 
	  it2!=it1->end(); it2++) {
	int x = it2->x;
	int y = it2->y;
	
	it2->x = round_double(cos_theta * x - sin_theta * y );
	it2->y = round_double(sin_theta * x + cos_theta * y );
	
      }
    
    rcone.setDirection(cos_theta,  sin_theta);
    rcone.init();
    _cones[i] = rcone;      
    
  }
}



int CircularVoting::nextConeRadius(int h, int r)
{
  //  return  r-1;
  if (h<=0 || r<0) {
    cerr << "Error in CircularVoting::nextConeRadius: null cone\n";
    return 0;
  }
  return (int)((double) h * tan(atan((double)r/(double)h) * 0.7 ));
}


void CircularVoting::vote()
{
  _sum = FloatImage(nx, ny);
  _mask = FloatImage(nx, ny);
  
  float* pmask = _mask.ptr();
  float* psum = _sum.ptr();
  
  vector<VPoint2D>::iterator voting_points_begin = _voting_points.begin();
  vector<VPoint2D>::iterator voting_points_end = _voting_points.end();
  
  for(vector<VPoint2D>::iterator it=voting_points_begin; 
      it!=voting_points_end; it++) {	
    it->pos = _sum.pos(it->x, it->y);
  }
  
  
  computeCones(_hmin, _hmax, _radius);
  
  for(vector<VPoint2D>::iterator it=voting_points_begin; 
      it!=voting_points_end; it++) {
    _cones[it->angIndex].vote(psum+it->pos, *it);
  }
  
  
  
  Image* roi;
  
  int loop = 0;
  
  int r = _radius;
  
  while (r>= 0) {
	
    //    cout << "Radius = " << r<< endl;
    computeCones(_hmin, _hmax, r);
    // _reg_type = BrightReg;
    
    if (_save_history) {
      roi = Roi::getSubImage(_sum, bw, bw, nx-bw-1, ny-bw-1);
      string vote_image_number = toString(r);
      if (r<10) vote_image_number = "0"+ vote_image_number;
      /* comment for saving 16bit ics file instead
	 writePGM(_prefix+"vote_"+vote_image_number+".pgm",
	 convertToByteImage(*roi, true));
      */
      image_io::writeICS(_prefix+"vote_"+vote_image_number,*roi);
      delete roi;
    }
    
    
    for(vector<VPoint2D>::iterator it=voting_points_begin; 
	it!=voting_points_end; it++) {	
      getMaxDirection(*it);    
    }    
    
    memcpy(pmask, psum, npix*sizeof(float));
    memset(psum, 0,  npix*sizeof(float));
    
    for(vector<VPoint2D>::iterator it=voting_points_begin; 
	it!=voting_points_end; it++) {
      _cones[it->angIndex].vote(psum+it->pos, pmask+it->pos, *it);
    }
    
    if (r<=0)
      break;
    
    r = nextConeRadius(_hmax, r);
    
  }
  
  computeCones(_hmin, _hmax, 0);
}




void CircularVoting::getMeanDirection(VPoint2D& vp)
{
  Cone2D& cone = _cones[vp.angIndex];
  Cone2D::iterator from = cone.begin();
  Cone2D::iterator to = cone.end();
  
  float* p = _sum.ptr() + vp.pos;
  float dx = 0;
  float dy = 0;
  int n = 0;
  float a;
  for(Cone2D::iterator it = from; it != to; it++) {
    n += it->size();
    for(ConePlane2D::iterator it1=it->begin(); it1!=it->end(); it1++) {
      a = p[it1->d];
      dx += a*it1->x;
      dy += a*it1->y;	
    }
  } 
  
  if (n<=0) return;
  
  dx /= n;
  dy /= n;
  
  dx = round_double(dx);
  dy = round_double(dy);
  
  float r = sqrt(dx*dx+dy*dy);
  if (r>epsilon) {
    vp.angIndex = computeAngleIndex(dx/r, dy/r);
  }
  
}

